#lang info

(define collection "sparcur")

(define deps '("base"
               "gui-lib"
               "gui-widget-mixins"
               "json-view"
               "orthauth"))

(define build-deps '())
